/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client.dataobject;

import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.data.AttributeGroupProtocolData;
import de.bsvrz.pua.prot.data.AttributeProtocolData;
import de.bsvrz.pua.prot.data.MultiColumnProtocolData;
import de.bsvrz.pua.prot.data.ProtocolData;
import de.bsvrz.pua.prot.processing.util.DataInformation;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ResultValue;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.*;


/**
 * Fasst die Daten eines Ergebnisdatensatzes zusammen. Diese Klasse bildet eine Zeile in einem Protokoll ab. Zum zugriff auf die einzelnen Werte
 * und Zustände bietet die Klasse verschiedene Getter-Methoden, die im Allgemeinen einen Spaltenindex als int-Wert akzeptieren.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolResultData implements ProtocolResult {

	private static final long serialVersionUID = -8268194346048847931L;

	/**
	 * Ergebnisdatensatz gehört zu der gesetzten Aggregation.
	 *
	 * @deprecated Ist immer {@link ProcessingInformation.ApplyAggregation#LISTE}, Ergebnisse von Aggregationen werden jetzt in einem {@link
	 *             de.bsvrz.pua.prot.client.dataobject.ProtocolAggregationResultData} verschickt.
	 */
	@Deprecated
	public static final ProcessingInformation.ApplyAggregation order = ProcessingInformation.ApplyAggregation.LISTE;

	/** Werte im Rohdatenformat. Stattdessen sollte getData(Datamodel, index) verwendet werden. */
	@Deprecated
	public ResultValue values[];

	/** Liste der Markierungen der Attribute. Stattdessen sollten die Methoden getDataKind(index) und getDataState(index) verwndet werden. */
	@Deprecated
	public DataInformation[] dataInformations;

	/** Liste der Datenzeitstempel */
	@Deprecated
	public long[] dataTimeStamps;

	/** Liste der Archivzeitstempel */
	@Deprecated
	public long[] archiveTimeStamps;

	/** Liste der Datensatzindizes */
	@Deprecated
	public long[] dataIndices;

	/** Status */
	@Deprecated
	public byte status;

	private int _intervalIndex;

	private ProtocolAdministrationData _adminData;
	
	private PuADataState[] _dataStates;


	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		StringBuffer sb = new StringBuffer();
		//nicht aufgefüllten Zeitstempel ausgeben (-> ist immer der größte Zeitstempel)
		long maxTime = getTimeStamp();

		sb.append("EDS ");
		Util.msToDate(sb, maxTime);
		sb.append(":\t");
		for(ResultValue value : values) {
			sb.append(value + "\t");
		}
		sb.append("Status: " + status + "%");
		return sb.toString();
	}


	@Override
	public String toString(final DataModel dataModel) {
		StringBuilder sb = new StringBuilder();
		//nicht aufgefüllten Zeitstempel ausgeben (-> ist immer der größte Zeitstempel)
		long maxTime = getTimeStamp();

		sb.append(String.format("%-28s\t", getIntervalIndex() + " " + Util.msToDate(maxTime)));

		final List<ProtocolData> dataList = getDataList(dataModel);
		for(int i = 0; i < dataList.size(); i++) {
			final ProtocolData value = dataList.get(i);
			PuADataState dataState = getDataState(i);
			if(dataState == PuADataState.DATA){
				sb.append(String.format("%28s\t", value.valueToString()));
			}
			else {
				sb.append(String.format("%28s\t", dataState + " (" + value.valueToString() + ")"));
			}
		}
		sb.append(getProgress()).append("%");
		return sb.toString();
	}


	/**
	 * Liefert den Zeitstempel zurück. Es wird der größte (-&gt; nicht aufgefüllte) Datenzeitstempel zurückgegeben.
	 *
	 * @return Zeitstempel.
	 */
	public long getTimeStamp() {
		long maxTime = Long.MIN_VALUE;
		for(long time : dataTimeStamps) {
			if(maxTime < time) maxTime = time;
		}
		return maxTime;
	}


	/**
	 * @param values            Werte.
	 * @param dataStates        Datenzustände
	 * @param dataInformations  Liste der Markierungen der Attribute. (Veraltet)
	 * @param dataTimeStamps    Liste der Datenzeitstempel.
	 * @param archiveTimeStamps Liste der Archivzeitstempel.
	 * @param dataIndices       Liste der Datensatzindizes.
	 * @param status            Status
	 * @param intervalIndex IntervalIndex
	 */
	public ProtocolResultData(
			ResultValue values[],
			final PuADataState[] dataStates,
			DataInformation dataInformations[],
			long dataTimeStamps[],
			long archiveTimeStamps[],
			long dataIndices[], byte status,
			int intervalIndex) {
		this.values = values;
		this.dataInformations = dataInformations;
		this.dataTimeStamps = dataTimeStamps;
		this.archiveTimeStamps = archiveTimeStamps;
		this.dataIndices = dataIndices;
		this.status = status;
		_intervalIndex = intervalIndex;
		_dataStates = dataStates;
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof ProtocolResultData) {
			ProtocolResultData other = (ProtocolResultData)object;

			return Arrays.equals(this.archiveTimeStamps, other.archiveTimeStamps) && Arrays.equals(this.dataIndices, other.dataIndices)
			       && Arrays.equals(this.dataInformations, other.dataInformations) && Arrays.equals(this.dataTimeStamps, other.dataTimeStamps)
			       && this.status == other.status && Arrays.equals(this.values, other.values);
		}
		else {
			return false;
		}
	}

	@Override
	public int hashCode() {
		return (int)((getTimeStamp() & 0xffff) + (_intervalIndex & 0x0f0f) * 47);
	}	

	/** @see de.bsvrz.pua.prot.client.dataobject.ProtocolResult#getType() */
	public int getType() {
		return ProtocolResult.PROTOCOL_DATA;
	}

	/**
	 * Wird intern aufgerufen, um clientseitig der Zeile den Spaltenkopf zu übergeben
	 * @param adminData Spaltenkopf
	 */
	@Deprecated
	public void setAdminData(final ProtocolAdministrationData adminData) {
		_adminData = adminData;
	}

	/**
	 * Gibt eine Liste der Data-Objekte zurück
	 * @param dataModel DataModel
	 * @return Liste mit Data-Objekten
	 */
	public List<ProtocolData> getDataList(final DataModel dataModel){
		final List<ProtocolData> result = new ArrayList<ProtocolData>(values.length);
		for(int i = 0; i < values.length; i++) {
			result.add(getData(dataModel, i) );
		}
		return result;
	}

	/**
	 * Gibt ein Data-Objekt zurück, dass die ganze Zeile beinhaltet
	 * @param dataModel DataModel
	 * @return Data-Objekt
	 */
	public ProtocolData getData(final DataModel dataModel) {
		final List<ProtocolData> list = getDataList(dataModel);
		return new MultiColumnProtocolData(list, _adminData.columns);
	}

	/**
	 * Gibt ein Data-Objekt zurück, dass einen Eintrag dieser Zeile enthält
	 * @param dataModel Datenmodell
	 * @param index Spaltenindex (0 &lt;= index &lt; getColumnCount())
	 * @return Data-Objekt
	 */
	public ProtocolData getData(final DataModel dataModel,final int index){
		final ResultValue value = values[index];
		final Column column = _adminData.columns[index];
		return createDataObject(dataModel, value, column);
	}

	/**
	 * Erstelt ein Data-Objekt
	 * @param dataModel Datenmodell
	 * @param value Wert des Datums
	 * @param column Spalte des Datums um Attributtyp usw. zu bestimmen
	 * @return Data-Objekt
	 */
	static ProtocolData createDataObject(final DataModel dataModel, final ResultValue value, final Column column) {
		final long referenceId = column.getReference();
		if(column.getType() == Column.ColumnType.DURATION_ATTRIBUTE){
			// Bei Zeitdauer-Attributen wird keine Attribut-Referenz mitgeliefert, es erscheint sinnvoll, hier einen passenden Attributtyp manuell zu wählen
			return new AttributeProtocolData(dataModel.getAttributeType("att.zeitDauer"), dataModel, value);
		}
		final SystemObject systemObject = dataModel.getObject(referenceId);
		if(systemObject instanceof Attribute) {
			final Attribute attribute = (Attribute)systemObject;
			return new AttributeProtocolData(attribute, dataModel, value);
		}
		if(systemObject instanceof AttributeType) {
			final AttributeType attributeType = (AttributeType)systemObject;
			return new AttributeProtocolData(attributeType, dataModel, value);
		}
		if(systemObject instanceof AttributeGroup) {
			final AttributeGroup attributeGroup = (AttributeGroup)systemObject;
			return new AttributeGroupProtocolData(attributeGroup, dataModel, value);
		}
		throw new IllegalArgumentException("Ungültige Attribut-ID im Spaltenkopf: " + referenceId + " (Systemobject: " + systemObject + ")");
	}

	/**
	 * Gibt den Zeitbereichs-Index zurück
	 * @return Der Index der angefragten Zeitbereiche
	 */
	public int getIntervalIndex() {
		return _intervalIndex;
	}

	/** 
	 * Gibt die Anzahl Spalten zurück
	 * @return die Anzahl Spalten
	 */
	public int getColumnCount(){
		return _adminData.getColumnCount();
	}

	/**
	 * Gibt den prozentuellen Fortschritt zurück. Dies ist die ungefähre prozentuale Position dieses Datensatzes im {@linkplain de.bsvrz.pua.prot.client.ProtocolResultStream}
	 * und kann beispielsweise dazu dienen, eine Fortschrittsanzeige beim Abruf eines Protokolls einzublenden.
	 * @return Fortschritt in Prozent
	 */
	public int getProgress() {
		return status;
	}

	/** 
	 * Gibt den Datenzeitstempel einer Zelle des Ergebnisdatensatzes zurück
	 * @param column Spalte des Ergebnisdatensatzes (0 &lt;= column &lt; getColumnCount())
	 * @return den Datenzeitstempel einer Zelle des Ergebnisdatensatzes 
	 * oder -1 falls der Datensatz keinen Zeitstempel besitzt (z. B. weil er nicht direkt einem Eingangsdatum entspricht)
	 */
	public long getDataTime(int column) {
		int did = _adminData.columnToDataId[column];
		if(did == -1) return -1;
		return dataTimeStamps[did];
	}

	/**
	 * Gibt den Archivzeitstempel einer Zelle des Ergebnisdatensatzes zurück
	 * @param column Spalte des Ergebnisdatensatzes (0 &lt;= column &lt; getColumnCount())
	 * @return den Archivzeitstempel einer Zelle des Ergebnisdatensatzes 
	 * oder -1 falls der Datensatz keinen Archivzeitstempel besitzt (z. B. weil er nicht direkt einem Eingangsdatum entspricht oder aus Onlinedaten gebildet wurde)
	 */
	public long getArchiveTime(int column) {
		int did = _adminData.columnToDataId[column];
		if(did == -1) return -1;
		return archiveTimeStamps[did];
	}

	/**
	 * Gibt den Datenindex einer Zelle des Ergebnisdatensatzes zurück
	 * @param column Spalte des Ergebnisdatensatzes (0 &lt;= column &lt; getColumnCount())
	 * @return den Datenindex einer Zelle des Ergebnisdatensatzes 
	 * oder -1 falls der Datensatz keinen Datenindex besitzt (z. B. weil er nicht direkt einem Eingangsdatum entspricht)
	 */
	public long getDataIndex(int column) {
		int did = _adminData.columnToDataId[column];
		if(did == -1) return -1;
		return dataIndices[did];
	}

	/**
	 * Gibt den (Fehler)zustand eines Datensatzes zurück.
	 * @param column Spalte des Ergebnisdatensatzes (0 &lt;= column &lt; getColumnCount())
	 * @return Ein PuADataState. Diese entsprechend weitesgehend der {@linkplain DataState}-Aufzählung, es gibt aber für PuA
	 * weitere Zustandswerte.
	 */
	public PuADataState getDataState(int column) {
		if(_dataStates == null) return PuADataState.NOT_AVAILABLE;
		return _dataStates[column];
	}

	/**
	 * Gibt die Archivdatenart zurück
	 * @param column Spalte des Ergebnisdatensatzes (0 &lt;= column &lt; getColumnCount())
	 * @return Archivdatenart
	 */
	public ArchiveDataKind getDataKind(int column) {
		int did = _adminData.columnToDataId[column];
		if(did == -1) return null;
		byte specialDataType = dataInformations[did].specialDataType;
		switch(specialDataType){
			case DataInformation.REQUESTED | DataInformation.DELAYED: return ArchiveDataKind.REQUESTED_DELAYED;
			case DataInformation.REQUESTED : return ArchiveDataKind.REQUESTED;
			case DataInformation.DELAYED: return ArchiveDataKind.ONLINE_DELAYED;
			default: return ArchiveDataKind.ONLINE;
		}
	}
}
