/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.data;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ResultValue;

import java.util.*;

/**
* Ein Datenobjekt, dass einen PuA-Ergebnisdatensatz mit dem zugehörigen Attributtyp verknüpft und damit einen einfachen Zugriff auf das Ergebnis bietet.<br><br>
 *
 * Bietet eine Kapselung der Protokolldaten in ein Interface, wie es Dav-Daten benutzen. Anders als Dav-Daten sind diese Daten nicht streng an die verwendete
 * Attributgruppe bzw. das verwendete Attribut gebunden (sofern überhaupt eines existiert). Je nach Anwendungsfall bieten sich daher auch Abfragen nach {@link
 * #isDefined()}, {@link #isPlain()} usw. an, auch wenn die Attributgruppe diese Daten eigentlich nicht unterstützen würde. Besonders wichtig ist dies bei
 * Fehlermeldungen von PuA (wie z.B. einer Division durch 0, oder einem "Keine Daten"-Datensatz), da hier die Meldung unabhängig vom Attributtyp über
 * {@link #asTextValue()} abgerufen werden kann.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public class AttributeProtocolData extends ProtocolData {

	private final AttributeType _attributeType;

	private final Attribute _attribute;

	private final ObjectLookup _dataModel;

	private final ResultValue _value;

	private boolean _forceArray = false;

	/**
	 * Enthält alle Kindattribute nach Index (für Daten-Arrays), sonst <code>null</code>
	 */
	private List<Data> _childrenByIndex = null;

	/**
	 * Enthält alle Kindattribute nach Bezeichnung (für Attributlisten), sonst <code>null</code>
	 */
	private Map<String, AttributeProtocolData> _childrenByString = null;

	/**
	 * Enthält die Werte dieses Arrays, sonst <code>null</code>
	 */
	private List<ResultValue> _arrayItems = null;

	public Data createUnmodifiableCopy() {
		throw new UnsupportedOperationException("Aus einzelnen Attributen kann kein eigenes Data-Objekt erstellt werden");
	}

	/**
	 * Gibt das in diesem Data-Objekt verwendete Datenmodell zurück
	 *
	 * @return DataModel
	 */
	ObjectLookup getDataModel() {
		return _dataModel;
	}

	/**
	 * Erstellt eine neue Instanz anhand eines Attributtyps
	 *
	 * @param attributeType Attributtyp
	 * @param dataModel     dataModel
	 * @param value         Wert von PuA
	 *
	 * @throws IllegalArgumentException Wenn Attributtyp nicht mit dem Wert von PuA kompatibel ist. Eine gewisse Toleranz ist erlaubt, aber z.B. in einer Liste
	 *                                  kann kein einzelnes atomares Datum sein.
	 */
	public AttributeProtocolData(final AttributeType attributeType, final ObjectLookup dataModel, final ResultValue value) throws IllegalArgumentException {
		_attribute = null;
		_attributeType = attributeType;
		_dataModel = dataModel;
		_value = value;
		initChildren();
	}

	/**
	 * Erstellt eine neue Instanz anhand eines Attributs
	 *
	 * @param attribute Attribut
	 * @param dataModel dataModel
	 * @param value     Wert von PuA
	 *
	 * @throws IllegalArgumentException Wenn Attribut nicht mit dem Wert von PuA kompatibel ist. Eine gewisse Toleranz ist erlaubt, aber z.B. in einer Liste
	 *                                  kann kein einzelnes atomares Datum sein.
	 */
	public AttributeProtocolData(final Attribute attribute, final ObjectLookup dataModel, final ResultValue value) throws IllegalArgumentException {
		_attribute = attribute;
		_attributeType = attribute.getAttributeType();
		_dataModel = dataModel;
		_value = value;
		initChildren();
	}

	private void initChildren() throws IllegalArgumentException {
		if(isList()) {
			if(_value.isAtomar()) {
				// throw new IllegalArgumentException(getName() + ": Der Attributtyp ist ein Listentyp, es sind aber nur atomare Daten vorhanden");
				return; // Bei einem "Keine Daten"-Ergebnis hier nicht zwingend ne Exception werfen, sodass man über asTextValue usw. noch an die eventuelle Fehlermeldung dran kommt.
			}
			AttributeListDefinition attributeListDefinition = (AttributeListDefinition)_attributeType;
			final List<ResultValue> children = _value.getChildren();
			final List<Attribute> attributes = attributeListDefinition.getAttributes();
			if(children.size() != attributes.size()) {
				throw new IllegalArgumentException(
						getName() + ": Die Anzahl der Unterattribute der Attributdefinition stimmt nicht mit der Anzahl der vorhandenen Daten überein"
				);
			}
			_childrenByString = new HashMap<String, AttributeProtocolData>(children.size());
			_childrenByIndex = new ArrayList<Data>(children.size());
			for(int i = 0, attributesSize = attributes.size(); i < attributesSize; i++) {
				final Attribute attribute = attributes.get(i);
				final AttributeProtocolData value = new AttributeProtocolData(attribute, _dataModel, children.get(i));
				_childrenByString.put(attribute.getName(), value);
				_childrenByIndex.add(value);
			}
		}
		else if(isArray()) {
			_arrayItems = _value.getChildren();
		}else{
			if(!_value.isAtomar()){
				_arrayItems = _value.getChildren();
				_forceArray = true;
			}
		}
	}

	public String getName() {
		if(_attribute != null) return _attribute.getName();
		return _attributeType.getName();
	}

	public AttributeType getAttributeType() {
		return _attributeType;
	}

	public boolean isDefined() {

		// Fehlerergebnisse und keine Ergebnisse sind generell nicht definiert
		final ExpressionResult value = _value.getValue();
		if(value != null) {
			if(value.getType() == ExpressionResult.ResultType.NONE
					|| value.getType() == ExpressionResult.ResultType.ERROR
					|| value.getType() == ExpressionResult.ResultType.NO_CHANGE) {
				return false;
			}
		}

		// Wenn in einer erwarteten Attributliste nur ein einzelner Wert drin ist, ist davon auszugehen, dass auch dies ein Fehlerwert ist.
		if(_attributeType instanceof AttributeListDefinition && _value.isAtomar()) return false;

		return _value.hasData();
	}

	@Override
	public boolean isNoChange() {
		ExpressionResult value = _value.getValue();
		return value != null && value.getType() == ExpressionResult.ResultType.NO_CHANGE;
	}

	public boolean isList() {
		return !isPlain() && _attributeType instanceof AttributeListDefinition && !isArray();
	}

	public boolean isArray() {
		return !isPlain() && ((_attribute != null && _attribute.isArray()) || _forceArray);
	}

	public boolean isPlain() {
		return _value.isAtomar();
	}

	public Data getItem(final String itemName) {
		if(!isList()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_LIST);
		return _childrenByString.get(itemName);
	}

	public TextValue asTextValue() {
		if(!isPlain()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
		if(isDefined()) {
			if(_attributeType instanceof ReferenceAttributeType) return asReferenceValue();
			if(_attributeType instanceof TimeAttributeType) return asTimeValue();
			if(_value.getValue().getType() == ExpressionResult.ResultType.LONG) return asScaledValue();
			if(_value.getValue().getType() == ExpressionResult.ResultType.DOUBLE) return asScaledValue();
			if(_value.getValue().getType() == ExpressionResult.ResultType.BOOL) return asScaledValue();
		}
		return new ProtocolDataTextValue(this);
	}


	public TimeValue asTimeValue() {
		if(!isPlain()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
		return new ProtocolDataTimeValue(this);
	}

	public NumberValue asScaledValue() {
		if(!isPlain()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
		return new ProtocolDataScaledValue(this);
	}

	public NumberValue asUnscaledValue() {
		if(!isPlain()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
		return new ProtocolDataUnscaledValue(this);
	}

	public ReferenceValue asReferenceValue() {
		if(!isPlain()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ATOMAR);
		return new ProtocolDataReferenceValue(this);
	}

	public Array asArray() {
		if(!isArray()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
		return new ProtocolDataArrayView(this);
	}

	public TextArray asTextArray() {
		if(!isArray()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
		return new ProtocolDataArrayView(this);
	}

	public TimeArray asTimeArray() {
		if(!isArray()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
		return new ProtocolDataArrayView(this);
	}

	public NumberArray asScaledArray() {
		if(!isArray()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
		return new ProtocolDataArrayView(this);
	}

	public NumberArray asUnscaledArray() {
		if(!isArray()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
		return new ProtocolUnscaledDataArrayView(this);
	}

	public ReferenceArray asReferenceArray() {
		if(!isArray()) throw new UnsupportedOperationException(getName() + MESSAGE_NOT_ARRAY);
		return new ProtocolDataArrayView(this);
	}

	/**
	 * Gibt die Array-Einträge als Rohdaten zurück. Wird z.B. für den Zugriff auf einzelne Elemente in {@link ProtocolDataArrayView} benutzt.
	 * @return Array-Einträge als Rohdaten
	 */
	List<ResultValue> getArrayItems() {
		return Collections.unmodifiableList(_arrayItems);
	}

	/**
	 * Gitbt den Wert als Rohdatum zurück. Wird benutzt um in {@link #asTextValue()} (bzw. dem darim erstellten Kapselungsobjekt) usw. auf den internen Wert zuzugreifen.
	 * @return Interner Wert dieses atomaren Datums. Bei Listen/Arraywerten wird null zurückgegeben.
	 */
	ExpressionResult getInternalValue() {
		return _value.getValue();
	}

	@Override
	public boolean isError() {
		// Fehlerergebnisse und keine Ergebnisse sind generell nicht definiert
		final ExpressionResult value = _value.getValue();
		if(value != null) {
			if(value.getType() == ExpressionResult.ResultType.ERROR) {
				return true;
			}
		}

		return false;

	}

	@Override
	protected List<Data> getChildrenList() {
		return Collections.unmodifiableList(_childrenByIndex);
	}

}
