/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;

import java.util.*;

/**
 * Zum Verwalten der Realen Attribute aus dem 'Spalten' Bereich. D.h. wenn erst der Filterbereich kommt, dann müssen die Attribute nicht in attributeInformation
 * abgelegt werden, und storeAggregationinformation muss diese Liste nicht nach passenden Attributen durchsuchen.
 * <p>
 * INFO Performance und Speicherverbrauch kann verbessert werden, wenn überprüft wird, welcher Bereich zuerst kommt.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class AttributeSemantics implements SemanticChecks {

	/** Zähler der angibt, wieviele reale Attribute und temporäre Attribute, die auf ein reales Attribut verweisen im Bereich Spalten definiert wurden. */
	private int _attributeCount = 0;

	/**
	 * Enthält reale Attribute, zu denen Aggregationsinformationen abgelegt wurden. Diese Attribute wurden aber noch nicht im Spaltenbereich definiert. (z.B:
	 * Attribut wird Aggregation zugewiesen, und Filter-Bereich liegt vor Spalten-Bereich.)
	 */
	private final HashMap<RealAttributeDescription, RealAttributeDescription> _aggregationInformation;

	/**
	 * Enthält reale Attribute, aus dem Spalten-Bereich oder aus Ausdrücken. Die Attribute können nicht in einer Hashtable abgelegt werden, da sie sich evtl. nur
	 * duch den Namen unterscheiden. INFO ggf. in eine Hastable<rad, liste> ändern.
	 */
	private final ArrayList<RealAttributeDescription> _attributeInformation;

	/** Konstruktor */
	public AttributeSemantics() {
		_aggregationInformation = new HashMap<RealAttributeDescription, RealAttributeDescription>();
		_attributeInformation = new ArrayList<RealAttributeDescription>();
	}

	/**
	 * Hinterlegt Aggregationsinformationen für die Attribute. Wenn für die übergebenen Attributbeschreibungen bereits weitere Informationen hinterlegt wurden,
	 * dann werden die Aggregationsinformationen dort eingetragen.
	 *
	 * @param attributes Attributbeschreibungen mit Aggregationsinformationen.
	 *
	 * @return Liefert das übergebene Feld zurück.
	 *
	 * @throws SemanticErrorException Falls zu einem Attribut unterschiedliche Aggregationsinformationen hinterlegt wreden.
	 * @see #storeAttributeInformation(RealAttributeDescription[])
	 */
	public RealAttributeDescription[] storeAggregationInformation(final RealAttributeDescription[] attributes) throws SemanticErrorException {
		for(final RealAttributeDescription attribute : attributes) {
			final RealAttributeDescription storedAttribute = _aggregationInformation.get(attribute);
			if(storedAttribute != null) {
				throw new SemanticErrorException(InterpreterErrorMessage.AGGREGATION_REMAPPING + attribute);
			}
			
			_aggregationInformation.put(attribute, attribute);
			//jetzt die Liste durchsuchen, ob die Aggregationsinformation irgendwo eingetragen werden kann.
			for(final RealAttributeDescription rad : _attributeInformation) {
				if(rad.equals(attribute)) rad.setAggregationClasses(attribute.getAggregationClasses());
			}
			
		}
		return attributes;
	}

	/**
	 * Hinterlegt Attributsbeschreibungen. Wenn für die übergebenen Attributsbeschreibungen Aggregationsinformationen vorliegen. dann werden diese
	 * Aggregationsinformationen die übergebenen Attribut einkopiert.
	 *
	 * @param attributes Attributbeschreibungen
	 *
	 * @return Attributbeschreibungen mit ggf. eingefügter Aggregationsinformation
	 *
	 * @see #storeAggregationInformation(RealAttributeDescription[])
	 */
	public RealAttributeDescription[] storeAttributeInformation(final RealAttributeDescription[] attributes) {
		for(final RealAttributeDescription attribute : attributes) {
			//in Liste aufnehmen
			_attributeInformation.add(attribute);
			//Überprüfen, ob es Aggregationsinformationen gibt. Diese dann eintragen.
			final RealAttributeDescription aggregation = _aggregationInformation.get(attribute);
			if(aggregation != null) attribute.setAggregationClasses(aggregation.getAggregationClasses());
		}
		return attributes;
	}


	/** Erhöht den Attributzähler. */
	public void incAttributeCount() {
		_attributeCount++;
	}

	/**
	 * Führt abschließende Semantische Überprüfungen durch. 
	 *
	 * @param pp            Wird nicht berücksichtigt
	 * @param processScript Keine Bedeutung
	 *
	 * @throws SemanticErrorException Es wird kein reales Attribut verwendet.
	 */
	public void cleanUp(final ProcessingParameter pp, final boolean processScript) throws SemanticErrorException {
	}
}
