/*
 * Copyright 2015 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.aggregations;

import de.bsvrz.pua.prot.util.ErrorMessageBuilder;
import de.bsvrz.pua.prot.util.ExpressionResult;
import de.bsvrz.pua.prot.util.ResultValue;

import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Supplier;

/**
 * Erweiterung der {@link ResultValue}-Klasse um zusätzliche Informationen für Aggregationen zu speichern (z.B. die Anzahl der aggregierten Werte zur Berechnung
 * von Mittelwerten. Bietet Methoden, um über strukturierte Datensätze zu iterieren.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @author Jonathan Haas
 */
@SuppressWarnings("rawtypes")
public class AggregationResultValue extends ResultValue {

	private static final long serialVersionUID = 8351161620443798933L;

	/** Aggregationszähler. */
	private int aggregationCount = 0;
	
	AbstractAggregation _aggregation;
	
	public static <T extends AggregationResultValue> T create(ResultValue src, Supplier<T> constructor, final AbstractAggregation<T> aggregation){
		T dest = constructor.get();
		dest._aggregation = aggregation;
		if(src.isAtomar()) {
			if(src.isSummable()) {
				dest.set(src.getValue().clone());
			}
			else {
				dest.set(new ExpressionResult());
			}
			if(dest.getValue().getType() != ExpressionResult.ResultType.NONE){
				dest.incAggregationCount();
			}
		}
		else {
			dest._value = null;
			ArrayList<ResultValue> children = new ArrayList<>(src.getChildren().size());
			for(ResultValue rv : src.getChildren()) {
				children.add(create(rv, constructor, aggregation));
			}
			dest._children = children;
		}
		return dest;
	}

	public void aggregate(final ResultValue right) {
		boolean leftAtomic = isAtomar();
		boolean rightAtomic = right.isAtomar();
		if(rightAtomic && right.getValue().getType() == ExpressionResult.ResultType.NONE){
			//Falls rechter Wert = none -> ignorieren. (passiert z.B. durch einen 'keine Daten' Datensatz)
			return;
		}
		if(leftAtomic != rightAtomic){
			set(ExpressionResult.error(ErrorMessageBuilder.illegalAggregation(this, right, _aggregation.getAggregationName())));
			return;
		}
		if(leftAtomic) {
			assert rightAtomic;
			if(_value.getType() == ExpressionResult.ResultType.NONE){
				set(right.getValue());
				if(getValue().getType() != ExpressionResult.ResultType.NONE) {
					incAggregationCount();
				}
			}
			else if(right.isSummable())	//Wert nur aggregieren, falls innerhalb des Wertebereiches (siehe TAnf S.22)
			{
				_aggregation.aggregate(this, right.getValue());
				incAggregationCount();
			}
		}
		else {
			//Es werden soviele Feldelemente wie möglich aggregiert. Werte, für die es
			//keinen Partner gibt, werden verworfen.
			List<ResultValue> rightChildren = right.getChildren();
			if(getChildren().size() > rightChildren.size()) truncChildren(rightChildren.size());
			List<ResultValue> children = getChildren();
			for(int i = 0; i < children.size(); i++) {
				((AggregationResultValue)children.get(i)).aggregate(rightChildren.get(i));;
			}
		}
	}

	/** Erhöht den Aggregationszähler für dieses Objekt. */
	public void incAggregationCount() {
		aggregationCount++;
	}

	/**
	 * Liefert den Wert des Aggregationszählers. Dieser Wert entspricht der Anzahl der aggregierten Werte
	 *
	 * @return Aggregationszähler.
	 */
	public int getAggregationCount() {
		return aggregationCount;
	}

	public void finish() {
		boolean atomic = isAtomar();
		if(atomic) {
			if(_value.getType() != ExpressionResult.ResultType.NONE
					&& _value.getType() != ExpressionResult.ResultType.ERROR) {
				_aggregation.finish(this);
			}
		}
		else {
			List<ResultValue> children = getChildren();
			for(ResultValue child : children) {
				((AggregationResultValue) child).finish();
			}
		}
	}

	private void writeObject(ObjectOutputStream oos) throws IOException {
		throw new NotSerializableException();
	}
}
