/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing.onlinebuffer;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.pua.prot.util.ReferenceQualifier;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.*;


/**
 * Verwaltet Anmeldungen zum Abgreifen von Online-Daten. Alle vom Set erzeugten Empfänger haben den gleichen Zeitbereich und die gleichen Optionen (Normal oder
 * Delta-Daten, nachgelieferte Daten). Es wird nur ein neuer Empfänger erzeugt, wenn es noch keinen Empfänger für die jeweilige Datenidentifikation gibt. Alle
 * Empfänger werden als einfahce Empfänger erzeugt (nicht als Senke). Der OnlineDataReceiver erzeugt genau einen Empfänger je Datenidentifikation.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class OnlineDataReceiverSet {

	/** Delta - Anfrage und nachgelieferte Daten */
	private ReceiveOptions _receiveOptions;

	/** Verbindung zum Datenverteiler */
	private ClientDavInterface _dav;

	/** Hashtable mit allen bisher erzeugten Empfängern. */
	private Hashtable<Qualifier, OnlineDataReceiverObject> _receivers;

	/** Liste mit den abzudeckenden Zeitbereichen */
	private List<Tuple<Long, Long>> _periods;

	/** Objekt das benachrichtigt wird, wenn neue Daten vorliegen. */
	private final NewDataCollector _newDataCollector;

	/**
	 * Erzeugt das Objekt.
	 *
	 * @param dav              Verbindung zum Datenverteiler
	 * @param receiveOptions   Anmeldeoptionen. (delta-Anfrage, nachgelieferte Daten). <b>Siehe auch Klassenbeschreibung für weitere Informationen.</b>
	 * @param periodList       Liste mit relevanten Zeitbereichen. Wird nicht verändert. <b>Siehe auch Klassenbeschreibung für weitere Informationen.</b>
	 * @param newDataCollector Objekt, das benachrichtigt wird, wenn neue Daten vorliegen.
	 *
	 * @throws FailureException Falls es bei der Kommunikation mit der Konfiguration zu einem Fehler kommt.
	 */
	public OnlineDataReceiverSet(ClientDavInterface dav, ReceiveOptions receiveOptions, List<Tuple<Long, Long>> periodList, NewDataCollector newDataCollector)
			throws FailureException {
		this._receiveOptions = receiveOptions;
		this._dav = dav;
		this._newDataCollector = newDataCollector;

		_periods = copyPeriods(periodList);
		// Hashtable mit erzeigten Receivern anlegen.
		_receivers = new Hashtable<Qualifier, OnlineDataReceiverObject>();
	}

	/**
	 * Gibt einen Empfänger für die übergebende Datenidentifikation zurück.
	 *
	 * @param qualifier Datenidentifikation.
	 *
	 * @return Empfänger für die übergebene Datenidentifikation.
	 *
	 * @throws FailureException Falls es bei der Kommunikation mit der Konfiguration zu einem Fehler kommt.
	 */
	public OnlineDataReceiver createReceiver(Qualifier qualifier) throws FailureException {
		OnlineDataReceiverObject receiver = _receivers.get(qualifier);
		if(receiver == null) {
			// Receiver exisitiert noch nicht
			if(qualifier instanceof ReferenceQualifier) {
				receiver = new OnlineDataReferenceReceiverObject(_dav, getReceiveOptions(), this, (ReferenceQualifier)qualifier, _newDataCollector);
			}
			else {
				receiver = new OnlineDataReceiverObject(this, qualifier, _newDataCollector);
			}
			_receivers.put(qualifier, receiver);
			receiver.subscribe(_receiveOptions);
		}
		// Objekt registrieren.
		return receiver.registerNewSubscriber();
	}

	/**
	 * Klont die übergebene Liste mit Zeitbereichen. INFO Sollte die Liste immutable machen, dann spart getPeriods Zeit
	 *
	 * @param periodList Liste; wird nicht verändert.
	 */
	private static List<Tuple<Long, Long>> copyPeriods(List<Tuple<Long, Long>> periodList) {
		ArrayList<Tuple<Long, Long>> periods = new ArrayList<Tuple<Long, Long>>();
		/*
		 * Liste mit Zeitbereichen klonen.
		 */
		for(Tuple<Long, Long> period : periodList) {
			periods.add(new Tuple<Long, Long>(new Long(period.first), new Long(period.last)));
		}
		return periods;
	}

	/**
	 * Liefert Verbindung zum Dav. Diese Verbindung darf vom einem Aufrufer nicht beendet ({@link ClientDavInterface#disconnect(boolean,java.lang.String)})
	 * werden!
	 *
	 * @return Verbindung zum Dav.
	 */
	protected ClientDavInterface getDav() {
		return _dav;
	}

	/**
	 * Liefert eine Liste mit Zeitbereichen. Diese Liste darf nicht verändert werden! INFO Immutable machen (mittels Collections.unmodifiableCollection).Tupel muss
	 * dann aber auch daran angepasst werden.
	 *
	 * @return Liste der relevanten Zeitbereiche.
	 */
	protected List<Tuple<Long, Long>> getPeriods() {
		return _periods;
	}

	/**
	 * Liefert die Empfangsoptionen.
	 *
	 * @return Empfangsoptionen.
	 */
	public ReceiveOptions getReceiveOptions() {
		return _receiveOptions;
	}

	/**
	 * Zeigt ob Daten für einen der Empfänger vorliegen.
	 *
	 * @return	<code>true</code> falls noch Daten für einen Empfänger vorliegen.
	 */
	public boolean dataAvailable() {
		for(OnlineDataReceiverObject receiver : _receivers.values()) {
			if(receiver.dataAvailable()) return true;
		}
		return false;
	}
}
