/*
 * Copyright 2015 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.pua.prot.client.dataobject.PuADataState;
import de.bsvrz.pua.prot.processing.util.ExpressionResultAndState;
import de.bsvrz.pua.prot.util.ExpressionResult.ResultType;

/**
 * Klasse zum Durchführen der arithmetischen Operationen auf long oder double Werten. Wird von {@link ExpressionTree} verwendet und
 * ermöglicht arithmetische Operationen auf {@link ExpressionResult} Objekten.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class ArithmeticOperationEx {

	/** Double Wert des linken Operanden */
	private ExpressionResultAndState _left;

	/** Double Wert des rechten Operanden */
	private ExpressionResultAndState _right;
	private ExpressionTree _leftTree = null;
	private ExpressionTree _rightTree = null;

	/**
	 * Setzt den linken Operanden
	 *
	 * @param r linker Operand
	 */
	public ArithmeticOperationEx(final ExpressionResultAndState r) {
		setLeft(r);
	}

	/**
	 * @param left  Linker Operand.
	 * @param right Rechter Operand.
	 */
	public ArithmeticOperationEx(final ExpressionResultAndState left, final ExpressionResultAndState right) {
		this(left);
		set(right);
	}

	/**
	 * Setzt den rechten Operand
	 *
	 * @param longN Rechter Operand
	 */
	public void set(final long longN) {
		this._right = new ExpressionResultAndState(new ExpressionResult(longN), PuADataState.COMPUTED);
	}

	/**
	 * Setzt den rechten Operand
	 *
	 * @param doubleN Rechter Operand
	 */
	public void set(final double doubleN) {
		this._right = new ExpressionResultAndState(new ExpressionResult(doubleN), PuADataState.COMPUTED);
	}

	/**
	 * Setzt den rechten Operand
	 *
	 * @param r Rechter Operand
	 */
	public void set(final ExpressionResultAndState r) {
		_right = r;
	}

	/**
	 * Prüft &lt;
	 *
	 * @return Ergebnis
	 */
	public ExpressionResultAndState less() {
		if(!canCompute()) return error(ExpressionTree.Operation.less);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() < _right.getResult().getDouble(), _left, _right);
	}

	private ExpressionResultAndState error(final ExpressionTree.Operation operation) {
		return ExpressionResultAndState.error(ErrorMessageBuilder.illegalOperation(operation, _leftTree, _rightTree, _left, _right));
	}

	private boolean canCompute() {
		return (_left.getType() == ResultType.LONG || _left.getType() == ResultType.DOUBLE)
				&& (_right.getType() == ResultType.LONG || _right.getType() == ResultType.DOUBLE);
	}

	/**
	 * Prüft &lt;=
	 *
	 * @return Ergebnis
	 */
	public ExpressionResultAndState lessEqual() {
		if(!canCompute()) return error(ExpressionTree.Operation.lessEqual);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() <= _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Prüft &gt;
	 *
	 * @return Ergebnis
	 */
	public ExpressionResultAndState greater() {
		if(!canCompute()) return error(ExpressionTree.Operation.greater);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() > _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Prüft &gt;=
	 *
	 * @return Ergebnis
	 */
	public ExpressionResultAndState greaterEqual() {
		if(!canCompute()) return error(ExpressionTree.Operation.greaterEqual);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() >= _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Prüft auf Gleichheit der Operanden.
	 *
	 * @return Ergebnis
	 */
	public ExpressionResultAndState equal() {
		if(!canCompute()) return error(ExpressionTree.Operation.equal);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() == _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Prüft auf Ungleichheit der Operanden.
	 *
	 * @return Ergebnis
	 */
	public ExpressionResultAndState notEqual() {
		if(!canCompute()) return error(ExpressionTree.Operation.notEqual);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() != _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Subtrahiert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResultAndState performMinusOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.subtract);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() - _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Addiert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResultAndState performPlusOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.add);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() + _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Multipliziert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResultAndState performMultOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.mult);
		return ExpressionResultAndState.deriveResult(_left.getResult().getDouble() * _right.getResult().getDouble(), _left, _right);
	}

	/**
	 * Dividiert die Operanden. Falls einer der Operanden vom Typ double ist, ist das Ergebnis auch vom Typ double.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResultAndState performDivOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.div);
		double a = _left.getResult().getDouble();
		double b = _right.getResult().getDouble();
		if(b == 0){
			return ExpressionResultAndState.error(ErrorMessageBuilder.divisionByZero(ExpressionTree.Operation.div, _leftTree, _rightTree, _left, _right));
		}
		return ExpressionResultAndState.deriveResult(a / b, _left, _right);
	}

	/**
	 * Ganzzahl-Division der Operanden. Das Ergebnis ist immer vom Typ long.
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResultAndState performDivIntOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.divInt);
		long a = (long) _left.getResult().getDouble();
		long b = (long) _right.getResult().getDouble();
		if(b == 0){
			return ExpressionResultAndState.error(ErrorMessageBuilder.divisionByZero(ExpressionTree.Operation.divInt, _leftTree, _rightTree, _left, _right));
		}
		return ExpressionResultAndState.deriveResult(a / b, _left, _right);
	}

	/**
	 * Modulo-Operation
	 *
	 * @return Ergebnis der Operation
	 */
	public ExpressionResultAndState performModOperation() {
		if(!canCompute()) return error(ExpressionTree.Operation.modulo);
		double a = _left.getResult().getDouble();
		double b = _right.getResult().getDouble();
		if(b == 0) {
			return ExpressionResultAndState.error(ErrorMessageBuilder.divisionByZero(ExpressionTree.Operation.modulo, _leftTree, _rightTree, _left, _right));
		}
		return ExpressionResultAndState.deriveResult(a % b, _left, _right);
	}

	/**
	 * Setzt den linken Operanden
	 *
	 * @param leftOperand Linker Operand
	 */
	public void setLeft(final ExpressionResultAndState leftOperand) {
		_left = leftOperand;
	}

	public void setLeftTree(final ExpressionTree leftTree) {
		_leftTree = leftTree;
	}

	public void setRightTree(final ExpressionTree rightTree) {
		_rightTree = rightTree;
	}

	/** @see Object#toString() */
	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder();

		sb.append(_left);

		sb.append(":");

		sb.append(_right);

		return sb.toString();
	}
}
